'''
Created on 03.04.2011
@author: achim brand
'''

from globals import FREESECT
from os import stat
from struct import pack, unpack
from datetime import datetime


'''
get file system information about file
'''
def GetFileSystemInfo(file, fileName, fileSize):
    print '----------------------------------------------------'
    print 'File system information about %s:' % fileName
    print '----------------------------------------------------'
    print 'Size: \t\t\t', fileSize, '(bytes)'
    print 'Creation Date: \t\t', datetime.utcfromtimestamp(stat(fileName).st_ctime).\
    strftime('%Y-%m-%d %H:%M:%S'), '(UTC)'
    print 'Last Modified Date: \t', datetime.utcfromtimestamp(stat(fileName).st_mtime).\
    strftime('%Y-%m-%d %H:%M:%S'), '(UTC)'
    print 'Last Access Date: \t', datetime.utcfromtimestamp(stat(fileName).st_atime).\
    strftime('%Y-%m-%d %H:%M:%S'), '(UTC)'
    print '----------------------------------------------------'
    print 'End of file system information.'
    print '----------------------------------------------------'
    pass
 

'''
conversion from little- to big-endian
cf. http://msdn.microsoft.com/en-us/library/dd926344%28v=office.12%29.aspx
-> Unless otherwise specified, all data in the Word binary file format is stored 
in little-endian format.
'''
def LittleToBig_2Bytes(bytes):
    return pack('>H', unpack('<H', bytes)[0]).encode('hex').upper()

def LittleToBig_4Bytes(bytes):
    return pack('>I', unpack('<I', bytes)[0]).encode('hex').upper()

def LittleToBig_8Bytes(bytes):
    return pack('>Q', unpack('<Q', bytes)[0]).encode('hex').upper()


'''
filetime is a 64-bit value representing the number of 100-nanosecond intervals since 
January 1, 1601 (UTC). UNIX timestamp is the number of seconds since January 1, 1970
Difference (in 100-ns): 10000000 * 60 * 60 * 24 * 365 * 369 + 10000000 * 60 * 60 * 
24 * 89 (89 leap days) = 116444736000000000
'''
def FiletimeToDatetime(filetime):
    return datetime.utcfromtimestamp((int(filetime,16) - 116444736000000000) / 10000000)


'''
conversion of a DTTM structure to Datetime
http://msdn.microsoft.com/en-us/library/dd943798%28v=office.12%29.aspx
http://docs.python.org/library/datetime.html?#strftime-strptime-behavior
'''
def DTTMTimeToDateTime(dttmtime):
    dttmtime = int(LittleToBig_4Bytes(dttmtime),16)
    if dttmtime == 0: return 0
    wdy = dttmtime & int('11100000000000000000000000000000',2)
    wdy = wdy >> 29
    yr = dttmtime & int('00011111111100000000000000000000',2)
    yr = (yr >> 20) + 1900
    mon = dttmtime & int('00000000000011110000000000000000',2)
    mon = mon >> 16
    dom = dttmtime & int('00000000000000001111100000000000',2)
    dom = dom >> 11
    hr = dttmtime & int('00000000000000000000011111000000',2)
    hr = hr >> 6
    mint = dttmtime & int('00000000000000000000000000111111',2)
    return datetime(yr,mon,dom,hr,mint).strftime('%Y-%m-%d %H:%M')


'''
Find last used array entry
'''
def LastUsed(endingOffset, file):
    i = 0
    file.seek(endingOffset,0)
    valueLastUsed = LittleToBig_4Bytes(file.read(4))
    while valueLastUsed == FREESECT:
            i = i + 1
            file.seek(endingOffset - (i * 4),0)
            valueLastUsed = LittleToBig_4Bytes(file.read(4))
    return int(valueLastUsed,16), endingOffset - (i * 4), i


'''
check if given offset is within the file
'''
def ReasonableOffset(sector, sectorSize, fileSize, name):
    if ((sector + 1) * sectorSize >= fileSize):
        print '-------------------------------'
        print '%s Offset is out of range.' % name
        print '-------------------------------'
        return 0
    else:
        return 1


'''
create a comma-separated string from the reviewrs array
'''
def StringReviewers(reviewers):
    i = 0
    strReviewers = ''
    for each in reviewers:
        strReviewers = strReviewers + reviewers[i] + ';' 
        i = i + 1
    return strReviewers.rstrip(';')


'''
get printer information
http://msdn.microsoft.com/en-us/library/dd926965%28v=office.12%29.aspx
'''
def GetPrinterInformation(file, WordDocumentOffset, xTableOffset, lcbPrDrvr):
    if lcbPrDrvr > 0:
        file.seek(WordDocumentOffset + 370,0)
        fcPrDrvr = int(LittleToBig_4Bytes(file.read(4)),16)
        PrDrvrOffset = xTableOffset + fcPrDrvr
        file.seek(PrDrvrOffset,0); PrDrvr = file.read(lcbPrDrvr).rstrip('\x00')
        (szPrinter, szPrPort, szPrDriver, szTruePrnName) = PrDrvr.split('\x00')
        print '----------------------------------------------------'
        print 'Printer Information:'
        print 'Offset to Printer Information: \t', PrDrvrOffset
        print '----------------------------------------------------'
        print 'Name: \t\t', szPrinter
        print 'Port: \t\t', szPrPort
        print 'Driver: \t', szPrDriver
        print 'Product Name: \t', szTruePrnName
        print '----------------------------------------------------'
        print 'End of Printer Information.'
        print '----------------------------------------------------'
    else:
        print 'No Printer Information found.'
    pass

