#!/usr/bin/env python

'''
Created on 03.04.2011
@author: achim brand
'''

from os import path, stat
from sys import exit, argv
from tools import GetFileSystemInfo, LittleToBig_2Bytes, LittleToBig_4Bytes, \
ReasonableOffset, StringReviewers, FiletimeToDatetime, GetPrinterInformation
from globals import CFBHEADER, WORDDOCUMENT, TABLE1, TABLE0, SUMMARYINFORMATION, DEFWIDENT
from FileSizeFAT import FileSizeFAT
from DopBaseData import GetDopBaseData
from SummaryInformationPropertySet import SummaryInformationPropertySet


def usage():
    print 'Usage: %s  wordfile.doc' % (path.basename(argv[0]))
    pass


def main():     
    if len(argv) != 2:
        usage()
        exit(0)
            
    try:
        fileName = argv[1]
        file = open(fileName, "rb")
        fileSize = stat(fileName).st_size
    except IOError:
        print 'Can not read', fileName
        exit(0)
    
      
    #print 'Actual File Size: \t\t%s \t\t\t (bytes)' % fileSize
    GetFileSystemInfo(file, fileName, fileSize)
    
    # read header signature 
    # header size is ALWAYS 512 bytes, independent from sector size
    # http://msdn.microsoft.com/en-us/library/dd941946%28v=PROT.13%29.aspx
    HeaderSignature = file.read(8)
    if HeaderSignature.encode('hex').upper() == CFBHEADER:        
        print 'Header Signature: \t\t0x%s \t (OLE compound file header)' % \
        HeaderSignature.encode('hex').upper()
    else:
        print '###################################################'
        print '# Header signature conforms not to an OLE header. #'
        print '###################################################'
        #exit(0)
        
    # read sector size; 
    # http://msdn.microsoft.com/en-us/library/dd941946%28v=PROT.13%29.aspx
    # A sector number can be converted into a byte offset into the file by using the 
    # following formula: (sector number + 1) x Sector Size. This implies that sector #0 of
    # the file begins at byte offset Sector Size, not at 0.
    # http://msdn.microsoft.com/en-us/library/dd942048%28v=PROT.13%29.aspx
    file.seek(30, 0); SectorShift = int(LittleToBig_2Bytes(file.read(2)), 16)
    SectorSize = 2 ** SectorShift
    #print 'Sector Size: \t\t', SectorSize, '\t\t\t (bytes)'
    
    # read number of FAT sectors
    file.seek(44, 0); NumFATSecs = int(LittleToBig_4Bytes(file.read(4)), 16)
    #print 'Number of FAT sectors: \t\t', NumFATSecs
    
    # check file size regarding FAT
    FileSizeFAT(NumFATSecs, SectorSize, fileSize, file)
    
            
    # read first directory sector location
    # http://msdn.microsoft.com/en-us/library/dd941946%28v=PROT.13%29.aspx
    file.seek(48, 0); DirSect1 = int(LittleToBig_4Bytes(file.read(4)), 16)
    if (ReasonableOffset(DirSect1, SectorSize, fileSize, 'DirSect1') == 0): exit(0)
    #print 'Offset for DirectoryEntries: \t', (DirSect1 + 1) * SectorSize    
   
    # find begin of WordDocument, SummaryInformation Directory Entries and check if there 
    # is a 1Table or 0Table Stream
    WordDocumentOffset = 0
    SummaryInformationOffset = 0
    xTableOffset = 0
    for i in range(100):
        # first directory entry is the root directory, therefore i+1
        file.seek((((DirSect1 + 1) * SectorSize) + ((i + 1) * 128)), 0) 
        if (file.read(26).encode('hex') == WORDDOCUMENT):            
            WordDocumentOffset = (((DirSect1 + 1) * SectorSize) + ((i + 1) * 128))
            file.seek(WordDocumentOffset + 116, 0)
            WordDocumentOffset = int(LittleToBig_4Bytes(file.read(4)), 16) 
            if (ReasonableOffset(WordDocumentOffset, SectorSize, fileSize, \
                                 'WordDocument') == 0): exit(0)
            WordDocumentOffset = (WordDocumentOffset + 1) * SectorSize
            #print 'Offset to Word Document: \t%s' % WordDocumentOffset
            
            # read fWhichTblStm (1Table or 0Table)
            # http://msdn.microsoft.com/en-us/library/dd944620%28v=office.12%29.aspx
            file.seek(WordDocumentOffset + 11, 0); fWhichTblStm = file.read(1)
            if ((int(fWhichTblStm.encode('hex'), 16) & 2) == 2): xTable = TABLE1
            else: xTable = TABLE0
            
        file.seek((((DirSect1 + 1) * SectorSize) + ((i + 1) * 128)), 0)
        if (file.read(40).encode('hex') == SUMMARYINFORMATION):            
            SummaryInformationOffset = (((DirSect1 + 1) * SectorSize) + ((i + 1) * 128))
            #print 'Summary Information Offset: ', SummaryInformationOffset       
            
        if (WordDocumentOffset > 0) and (SummaryInformationOffset > 0) : break
    
    if (WordDocumentOffset == 0) or (SummaryInformationOffset == 0):
        print '#############################'
        print '# Needed offsets not found. #'
        print '# Program is terminated.    #'
        print '#############################'
        exit(0)
                 
    # find begin of 1Table or 0Table Directory Entries
    for i in range(100):
        # first directory entry is the root directory, therefore i+1
        file.seek((((DirSect1 + 1) * SectorSize) + ((i + 1) * 128)), 0)    
        if (file.read(14).encode('hex') == xTable[1]):
            xTableOffset = (((DirSect1 + 1) * SectorSize) + ((i + 1) * 128))
            file.seek(xTableOffset + 116, 0)
            xTableOffset = int(LittleToBig_4Bytes(file.read(4)), 16)
            if (ReasonableOffset(xTableOffset, SectorSize, fileSize, \
                                 xTable[0]) == 0): exit(0)
            xTableOffset = (xTableOffset + 1) * SectorSize
            print 'Offset to %s: \t\t%s' % (xTable[0], xTableOffset)
            break
    
    # read wIdent
    # http://msdn.microsoft.com/en-us/library/dd944620%28v=office.12%29.aspx
    file.seek(WordDocumentOffset, 0); wIdent = LittleToBig_2Bytes(file.read(2))
    if wIdent == DEFWIDENT:
        print 'File Identification: \t\t0x%s\t\t\t (Word Binary File)' % wIdent
    else:
        print '############################################'
        print '# File conforms not to a Word binary file. #'
        print '############################################'
        #exit(0)
    
    # read data from Dop (Document properties); 
    # fcDop specifies an offset in the Table Stream, where the Dop begins
    # http://msdn.microsoft.com/en-us/library/dd905534%28v=office.12%29.aspx
    # http://msdn.microsoft.com/en-us/library/dd944942%28v=office.12%29.aspx
    file.seek(WordDocumentOffset + 402, 0)
    fcDop = int(LittleToBig_4Bytes(file.read(4)), 16)
    DopBaseOffset = xTableOffset + fcDop
    GetDopBaseData(DopBaseOffset, file)
    
    # get offset for summary information property set
    file.seek(SummaryInformationOffset + 116, 0)
    SummaryInformationpropertySetOffset = int(LittleToBig_4Bytes(file.read(4)), 16)
    if (ReasonableOffset(SummaryInformationpropertySetOffset, SectorSize, fileSize, \
                         'SummaryInformationpropertySet') == 0): exit(0)
    SummaryInformationpropertySetOffset = (SummaryInformationpropertySetOffset + 1) \
    * SectorSize
    #print 'SummaryInformationPropertySet Offset: ', SummaryInformationpropertySetOffset
    
    # read summary information property set (.doc metadata)
    # http://msdn.microsoft.com/en-us/library/dd942089%28v=PROT.13%29.aspx 
    # http://msdn.microsoft.com/en-us/library/dd942543%28v=PROT.13%29.aspx
    
    # Offset0 is 44 bytes after SummaryInformationpropertySetOffset
    file.seek(SummaryInformationpropertySetOffset + 44, 0)
    Offset0 = int(LittleToBig_4Bytes(file.read(4)), 16)
    #print 'Offset0: ', Offset0 
    
    # beginning of property set 0
    # http://msdn.microsoft.com/en-us/library/dd942089%28v=PROT.13%29.aspx
    propertySetOffset = SummaryInformationpropertySetOffset + Offset0 
    #print 'Property Set Offset: ', propertySetOffset
    
    # create SummaryInformationPropertySet object 
    sips = SummaryInformationPropertySet(propertySetOffset, file)
    # print SIPS
    sips.printSIPS()
    
    print '----------------------------------------------------'
    print 'Miscellaneous Metadata.'
    print '----------------------------------------------------'
        
    # read reviewers
    # read lcbSttbfRMark (size of SttbfRMark)  and fcSttbfRMark (offset to reviewers) 
    # from FibRgFcLcb97
    # http://msdn.microsoft.com/en-us/library/dd905534%28v=office.12%29.aspx
    # http://msdn.microsoft.com/en-us/library/dd953637%28v=office.12%29.aspx
    # http://msdn.microsoft.com/en-us/library/dd906744%28v=office.12%29.aspx
    file.seek(WordDocumentOffset + 566, 0)
    lcbSttbfRMark = int(LittleToBig_4Bytes(file.read(4)), 16)
    #print 'lcbSttbfRMark: ', lcbSttbfRMark
    if (lcbSttbfRMark > 0):
        file.seek(WordDocumentOffset + 562, 0)
        fcSttbfRMark = int(LittleToBig_4Bytes(file.read(4)), 16)
        #print 'fcSttbfRMark: ', fcSttbfRMark
        file.seek(xTableOffset + fcSttbfRMark + 2, 0)
        numReviewers = int(LittleToBig_2Bytes(file.read(2)), 16)
        offsetReviewers = xTableOffset + fcSttbfRMark + 6
        print 'Number of Reviewers: \t\t', numReviewers - 1
        print 'Offset to Reviewers: \t\t', offsetReviewers           
        reviewers = []
        for i in range(numReviewers):
            file.seek(offsetReviewers, 0)
            lenReviewer = int(LittleToBig_2Bytes(file.read(2)), 16)
            #print 'Length of reviewers name:', lenReviewer
            # 2 * lenReviewer -> data fields in STTB contain extended (2-byte) characters
            reviewers.append(file.read(2 * lenReviewer).replace('\x00', ''))
            offsetReviewers = offsetReviewers + 2 + 2 * lenReviewer;
            #print 'Offset to reviewer:', offsetReviewers
        if ((numReviewers - 1) > 0): 
            print 'Reviewers: \t\t\t', StringReviewers(reviewers[1:])
    else:
        print 'No Reviewers found.'
        
    # read last saved date from FibRgFcLcb97
    # http://msdn.microsoft.com/en-us/library/dd905534%28v=office.12%29.aspx
    file.seek(WordDocumentOffset + 850, 0); dwLowDateTime = file.read(4)
    if (int(dwLowDateTime.encode('hex'), 16) != 0): 
        dwLowDateTime = LittleToBig_4Bytes(dwLowDateTime)
    file.seek(WordDocumentOffset + 854, 0); dwHighDateTime = file.read(4)
    if (int(dwHighDateTime.encode('hex'), 16) != 0): 
        dwHighDateTime = LittleToBig_4Bytes(dwHighDateTime)
    if (int(dwHighDateTime.encode('hex'), 16) != 0) or \
    (int(dwHighDateTime.encode('hex'), 16) != 0):
        print 'Last Save Date (from FIB): \t%s (UTC)' \
        % FiletimeToDatetime(dwHighDateTime + dwLowDateTime)
    
    # read printer information; fcPrDrvr and lcbPrDrvr
    #http://msdn.microsoft.com/en-us/library/dd905534%28v=office.12%29.aspx
    file.seek(WordDocumentOffset + 374, 0)
    lcbPrDrvr = int(LittleToBig_4Bytes(file.read(4)), 16)
    GetPrinterInformation(file, WordDocumentOffset, xTableOffset, lcbPrDrvr)
        
    file.close()
    print '\n--- end of script ---\n'
    
    pass


if __name__ == '__main__':
    main()